//----------------------------------------------------------------------------
//
// File:        screenio.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#if defined ( __OS2__ )
    #define INCL_SUB
    #define INCL_NOPMAPI
    #include <os2.h>
    #include <conio.h>
#elif defined ( __WIN32__ )
    #include <conio.h>
    #include <windows.h>
#elif defined ( __GNUC__ )
    #include <memory.h>
    #include <signal.h>
    #include <stdio.h>
    #include <stdlib.h>
    #include <sys/time.h>
    #include <termios.h>
    #include <unistd.h>
#else
    #error "This code must be compiled for either OS/2, Win32, or Linux"
#endif
#include "common.hpp"
#include "screenio.hpp"

#define toHex(x)	( char ) (((x) > 9 ) ? (x) + 'A' - 10 : (x) + '0' )

int curX = 1;
int curY = 1;

#if defined ( __OS2__ )

static HVIO  hVio = 0;

void GotoXY ( int x, int y )
{
    curX = x;
    curY = y;
}

void PutXY ( int x, int y, char *ptr, int length )
{
    VioWrtCharStr ( ptr, length, y-1, x-1, hVio );
    curX = x + length;
    curY = y;
}

void Put ( char *ptr, int length )
{
    VioWrtCharStr ( ptr, length, curY-1, curX-1, hVio );
    curX += length;
}

#elif defined ( __WIN32__ )

static HANDLE hOutput = GetStdHandle ( STD_OUTPUT_HANDLE );

static COORD  currentPos;
static CONSOLE_CURSOR_INFO cursorInfo;
static CONSOLE_SCREEN_BUFFER_INFO screenInfo;

void SaveConsoleSettings ()
{
    GetConsoleCursorInfo ( hOutput, &cursorInfo );
    GetConsoleScreenBufferInfo ( hOutput, &screenInfo );
}

void RestoreConsoleSettings ()
{
    SetConsoleCursorInfo ( hOutput, &cursorInfo );
    SetConsoleCursorPosition ( hOutput, screenInfo.dwCursorPosition );
}

#include <stdio.h>

int GetKey ()
{
    int key = 0;
    UCHAR *ptr = ( UCHAR * ) &key;
    int ch = getch ();

    if ( ch == 0xE0 ) {
        *ptr++ = 0x1B;
        *ptr++ = 0x5B;
        ch = getch ();
        switch ( ch ) {
            case 0x48 : ch = 0x41;  break;
            case 0x50 : ch = 0x42;  break;
            case 0x4D : ch = 0x43;  break;
            case 0x4B : ch = 0x44;  break;
        }
    }
    *ptr++ = ch;

    return key;
}

bool KeyPressed ()
{
    return ( kbhit () != 0 ) ? true : false;
}

void HideCursor ()
{
    CONSOLE_CURSOR_INFO info = cursorInfo;
    info.bVisible = FALSE;
    SetConsoleCursorInfo ( hOutput, &info );
}

void ShowCursor ()
{
    CONSOLE_CURSOR_INFO info = cursorInfo;
    info.bVisible = TRUE;
    SetConsoleCursorInfo ( hOutput, &info );
}

void ClearScreen ()
{
    DWORD dwActual;
    COORD origin;
    origin.X = 0;
    origin.Y = 0;
    FillConsoleOutputCharacter ( hOutput, ' ', screenInfo.dwSize.X * screenInfo.dwSize.Y, origin, &dwActual );

    screenInfo.dwCursorPosition.X = 0;
    screenInfo.dwCursorPosition.Y = 0;
}

void GotoXY ( int x, int y )
{
    curX = x;
    curY = y;
    if ( cursorInfo.bVisible == TRUE ) {
        COORD cursor;
        cursor.X = curX - 1;
        cursor.Y = curY - 1;
        SetConsoleCursorPosition ( hOutput, cursor );
    }
}

void PutXY ( int x, int y, char *ptr, int length )
{
    DWORD count;
    currentPos.X = ( SHORT ) ( x - 1 );
    currentPos.Y = ( SHORT ) ( y - 1 );
    WriteConsoleOutputCharacter ( hOutput, ptr, length, currentPos, &count );
    curX = x + length;
    curY = y;
}

void Put ( char *ptr, int length )
{
    DWORD count;
    currentPos.X = ( SHORT ) ( curX - 1 );
    currentPos.Y = ( SHORT ) ( curY - 1 );
    WriteConsoleOutputCharacter ( hOutput, ptr, length, currentPos, &count );
    curX += length;
}

#elif defined ( __GNUC__ )

static termios stored;
static termios term_getch;
static termios term_kbhit;
static int lastChar;
static int keyhit;
static bool cursor_visible = true;

extern char *strupr ( char *ptr );

int GetKey ()
{
    int retVal = lastChar;
    lastChar = 0;

    if ( keyhit == 0 ) {
        tcsetattr ( 0, TCSANOW, &term_getch );
        keyhit = read ( STDIN_FILENO, &retVal, sizeof ( retVal ));
    }

    keyhit = 0;

    return retVal;
}

bool KeyPressed ()
{
    if ( keyhit == 0 ) {
        tcsetattr ( 0, TCSANOW, &term_kbhit );
        keyhit = read ( STDIN_FILENO, &lastChar, sizeof ( lastChar ));
    }

    return ( keyhit != 0 ) ? true : false;
}

void SaveConsoleSettings ();
void RestoreConsoleSettings ();

void SignalHandler ( int signal )
{
    struct sigaction sa;
    memset ( &sa, 0, sizeof ( sa ));
    sa.sa_handler = SIG_DFL;
    switch ( signal ) {
        case SIGABRT :
            RestoreConsoleSettings ();
            break;
        case SIGINT :
            RestoreConsoleSettings ();
            printf ( "\r\n" );
            exit ( -1 );
            break;
        case SIGTSTP :
            RestoreConsoleSettings ();
            printf ( "\r\n" );
            sigaction ( SIGTSTP, &sa, NULL );
            kill ( getpid (), SIGTSTP );
            break;
        case SIGCONT :
            SaveConsoleSettings ();
            break;
    }
}

void SaveConsoleSettings ()
{
    tcgetattr ( 0, &stored );
    memcpy ( &term_getch, &stored, sizeof ( struct termios ));
    // Disable echo
    term_getch.c_lflag &= ~ECHO;
    // Disable canonical mode, and set buffer size to 1 byte
    term_getch.c_lflag &= ~ICANON;
    term_getch.c_cc[VMIN] = 1;

    memcpy ( &term_kbhit, &term_getch, sizeof ( struct termios ));
    term_kbhit.c_cc[VTIME] = 0;
    term_kbhit.c_cc[VMIN] = 0;

    HideCursor ();

    struct sigaction sa;
    memset ( &sa, 0, sizeof ( sa ));
    sa.sa_handler = SignalHandler;
    sigaction ( SIGABRT, &sa, NULL );
    sigaction ( SIGINT,  &sa, NULL );
    sigaction ( SIGTSTP, &sa, NULL );
    sigaction ( SIGCONT, &sa, NULL );
}

void RestoreConsoleSettings ()
{
    struct sigaction sa;
    memset ( &sa, 0, sizeof ( sa ));
    sa.sa_handler = SIG_DFL;
    sigaction ( SIGABRT, &sa, NULL );
    sigaction ( SIGINT,  &sa, NULL );
    sigaction ( SIGTSTP, &sa, NULL );
    sigaction ( SIGCONT, &sa, NULL );

    tcsetattr ( 0, TCSANOW, &stored );

    ShowCursor ();

    printf ( "\n" );
}

void ClearScreen ()
{
    printf ( "\033[2J" );
    fflush ( stdout );
}

void HideCursor ()
{
    if ( cursor_visible == true ) {
        printf ( "\033[?25l" );
        fflush ( stdout );
        cursor_visible = false;
    }
}

void ShowCursor ()
{
    if ( cursor_visible == false ) {
        printf ( "\033[?25h" );
        fflush ( stdout );
        cursor_visible = true;
    }
}

void GotoXY ( int x, int y )
{
    printf ( "\033[%d;%dH", y, x );
    fflush ( stdout );
    curX = x;
    curY = y;
}


void PutXY ( int x, int y, char *ptr, int length )
{
    GotoXY ( x, y );
    Put ( ptr, length );
}

void Put ( char *ptr, int length )
{
    static char buffer [80];
    if ( length > 80 ) length = 80;
    strncpy ( buffer, ptr, length );
    for ( int i = 0; i < length; i++ ) {
        if (( buffer [i] < 32 ) || ( buffer [i] > 126 )) {
            buffer [i] = '.';
        }
    }

    printf ( "%*.*s", length, length, buffer );
    fflush ( stdout );
    curX += length;
}

#endif

void outByte ( UCHAR val )
{
    char buffer [ 3 ], *ptr = &buffer[3], digit;
    *--ptr = '\0';
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    *--ptr = toHex ( val );
    Put ( ptr, 2 );
}

void outWord ( USHORT val )
{
    char buffer [5], *ptr = &buffer[5], digit;
    *--ptr = '\0';
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    *--ptr = toHex (( char ) val );
    Put ( ptr, 4 );
}

void outLong ( ULONG val )
{
    char buffer [9], *ptr = &buffer[9], digit;
    *--ptr = '\0';
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    digit = ( char ) ( val & 0x0F );    val >>= 4;
    *--ptr = toHex ( digit );
    *--ptr = toHex (( char ) val );
    Put ( ptr, 8 );
}
